@@ -23,11 +23,13 @@ import (
 	"io"
 	"os/exec"
 	"runtime"
+	"strings"
 	"sync"
 	"syscall"
 	"time"
 
-	. "github.com/containerd/containerd"
+	"github.com/containerd/containerd/v2/client"
+	"github.com/containerd/plugin"
 )
 
 type daemon struct {
@@ -55,31 +57,46 @@ func (d *daemon) start(name, address string, args []string, stdout, stderr io.Wr
 	return nil
 }
 
-func (d *daemon) waitForStart(ctx context.Context) (*Client, error) {
+func (d *daemon) waitForStart(ctx context.Context) (*client.Client, error) {
 	var (
-		client  *Client
-		serving bool
-		err     error
-		ticker  = time.NewTicker(500 * time.Millisecond)
+		clientInstance *client.Client
+		serving        bool
+		err            error
+		ticker         = time.NewTicker(500 * time.Millisecond)
 	)
 	defer ticker.Stop()
 
 	for {
 		select {
 		case <-ticker.C:
-			client, err = New(d.addr)
+			clientInstance, err = client.New(d.addr)
 			if err != nil {
 				continue
 			}
-			serving, err = client.IsServing(ctx)
+			serving, err = clientInstance.IsServing(ctx)
 			if !serving {
-				client.Close()
+				clientInstance.Close()
 				if err == nil {
 					err = errors.New("connection was successful but service is not available")
 				}
 				continue
 			}
-			return client, err
+			resp, perr := clientInstance.IntrospectionService().Plugins(ctx)
+			if perr != nil {
+				return nil, fmt.Errorf("failed to get plugin list: %w", perr)
+			}
+			var loadErr error
+			for _, p := range resp.Plugins {
+				if p.InitErr != nil && !strings.Contains(p.InitErr.Message, plugin.ErrSkipPlugin.Error()) {
+					pluginErr := fmt.Errorf("failed to load %s.%s: %s", p.Type, p.ID, p.InitErr.Message)
+					loadErr = errors.Join(loadErr, pluginErr)
+				}
+			}
+			if loadErr != nil {
+				return nil, loadErr
+			}
+
+			return clientInstance, err
 		case <-ctx.Done():
 			return nil, fmt.Errorf("context deadline exceeded: %w", err)
 		}
