@@ -31,6 +31,7 @@ import (
 	"github.com/containerd/containerd/v2/core/snapshots"
 	"github.com/containerd/containerd/v2/core/snapshots/storage"
 	"github.com/containerd/containerd/v2/core/snapshots/testsuite"
+	"github.com/containerd/containerd/v2/internal/userns"
 	"github.com/containerd/containerd/v2/pkg/testutil"
 	"github.com/containerd/containerd/v2/plugins/snapshots/overlay/overlayutils"
 	"github.com/opencontainers/runtime-spec/specs-go"
@@ -138,6 +139,65 @@ func testOverlayCommit(t *testing.T, newSnapshotter testsuite.SnapshotterFunc) {
 	if err := o.Commit(ctx, "base", key); err != nil {
 		t.Fatal(err)
 	}
+
+	activeParent := "active-parent"
+	if _, err := o.Prepare(ctx, activeParent, ""); err != nil {
+		t.Fatal(err)
+	}
+
+	// test rebase
+	testCases := []struct {
+		oldParent string
+		newParent string
+		expError  bool
+	}{
+		{
+			oldParent: "",
+			newParent: "",
+			expError:  false,
+		},
+		{
+			oldParent: "",
+			newParent: "base",
+			expError:  false,
+		},
+		{
+			oldParent: "base",
+			newParent: "base",
+			expError:  false,
+		},
+		{
+			oldParent: "base",
+			newParent: "",
+			expError:  false,
+		},
+		{
+			oldParent: "base",
+			newParent: "new",
+			expError:  true,
+		},
+		{
+			oldParent: "",
+			newParent: activeParent,
+			expError:  true,
+		},
+	}
+	for i, tc := range testCases {
+		key := fmt.Sprintf("/tmp/test-%d", i)
+		name := fmt.Sprintf("test-%d", i)
+		_, err := o.Prepare(ctx, key, tc.oldParent)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if err := o.Commit(ctx, name, key, snapshots.WithParent(tc.newParent)); err != nil {
+			if !tc.expError {
+				t.Fatal(err)
+			}
+			t.Logf("expected error received: %v", err)
+		} else if tc.expError {
+			t.Fatal("expected error but commit succeeded")
+		}
+	}
 }
 
 func testOverlayOverlayMount(t *testing.T, newSnapshotter testsuite.SnapshotterFunc) {
@@ -199,68 +259,280 @@ func testOverlayOverlayMount(t *testing.T, newSnapshotter testsuite.SnapshotterF
 }
 
 func testOverlayRemappedBind(t *testing.T, newSnapshotter testsuite.SnapshotterFunc) {
-	var (
-		opts   []snapshots.Opt
-		mounts []mount.Mount
-	)
+	for _, test := range []struct {
+		idMap        userns.IDMap
+		snapOpt      func(idMap userns.IDMap) snapshots.Opt
+		expUID       uint32
+		expGID       uint32
+		expUIDMntOpt string
+		expGIDMntOpt string
+	}{
+		{
+			idMap: userns.IDMap{
+				UidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      666,
+						Size:        65536,
+					},
+				},
+				GidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      666,
+						Size:        65536,
+					},
+				},
+			},
+			snapOpt: func(idMap userns.IDMap) snapshots.Opt {
+				return containerd.WithRemapperLabels(
+					idMap.UidMap[0].ContainerID, idMap.UidMap[0].HostID,
+					idMap.GidMap[0].ContainerID, idMap.GidMap[0].HostID,
+					idMap.UidMap[0].Size,
+				)
+			},
+			expUID:       666,
+			expGID:       666,
+			expUIDMntOpt: "uidmap=0:666:65536",
+			expGIDMntOpt: "gidmap=0:666:65536",
+		},
+		{
+			idMap: userns.IDMap{
+				UidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      666,
+						Size:        1000,
+					},
+					{
+						ContainerID: 1000,
+						HostID:      6666,
+						Size:        64536,
+					},
+				},
+				GidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      888,
+						Size:        1000,
+					},
+					{
+						ContainerID: 1000,
+						HostID:      8888,
+						Size:        64536,
+					},
+				},
+			},
+			snapOpt: func(idMap userns.IDMap) snapshots.Opt {
+				return containerd.WithUserNSRemapperLabels(idMap.UidMap, idMap.GidMap)
+			},
+			expUID:       666,
+			expGID:       888,
+			expUIDMntOpt: "uidmap=0:666:1000,1000:6666:64536",
+			expGIDMntOpt: "gidmap=0:888:1000,1000:8888:64536",
+		},
+	} {
+		var (
+			opts   []snapshots.Opt
+			mounts []mount.Mount
+		)
 
-	ctx := context.TODO()
-	root := t.TempDir()
-	o, _, err := newSnapshotter(ctx, root)
-	if err != nil {
-		t.Fatal(err)
+		ctx := context.TODO()
+		root := t.TempDir()
+		o, _, err := newSnapshotter(ctx, root)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		if sn, ok := o.(*snapshotter); !ok || !sn.remapIDs {
+			t.Skip("overlayfs doesn't support idmapped mounts")
+		}
+
+		opts = append(opts, test.snapOpt(test.idMap))
+
+		key := "/tmp/test"
+		if mounts, err = o.Prepare(ctx, key, "", opts...); err != nil {
+			t.Fatal(err)
+		}
+
+		bp := getBasePath(ctx, o, root, key)
+		expected := []string{test.expUIDMntOpt, test.expGIDMntOpt, "rw", "rbind"}
+
+		checkMountOpts := func() {
+			if len(mounts) != 1 {
+				t.Errorf("should only have 1 mount but received %d", len(mounts))
+			}
+
+			if len(mounts[0].Options) != len(expected) {
+				t.Errorf("expected %d options, but received %d", len(expected), len(mounts[0].Options))
+			}
+
+			m := mounts[0]
+			for i, v := range expected {
+				if m.Options[i] != v {
+					t.Errorf("mount option %q is not valid, expected %q", m.Options[i], v)
+				}
+			}
+
+			st, err := os.Stat(filepath.Join(bp, "fs"))
+			if err != nil {
+				t.Errorf("failed to stat %s", filepath.Join(bp, "fs"))
+			}
+
+			if stat, ok := st.Sys().(*syscall.Stat_t); !ok {
+				t.Errorf("incompatible types after stat call: *syscall.Stat_t expected")
+			} else if stat.Uid != test.expUID || stat.Gid != test.expGID {
+				t.Errorf("bad mapping: expected {uid: %d, gid: %d}; real {uid: %d, gid: %d}", test.expUID, test.expGID, int(stat.Uid), int(stat.Gid))
+			}
+		}
+		checkMountOpts()
+
+		expected[2] = "ro"
+		if err = o.Commit(ctx, "base", key, opts...); err != nil {
+			t.Fatal(err)
+		}
+		if mounts, err = o.View(ctx, key, "base", opts...); err != nil {
+			t.Fatal(err)
+		}
+		bp = getBasePath(ctx, o, root, key)
+		checkMountOpts()
+
+		key = "/tmp/test1"
+		if mounts, err = o.Prepare(ctx, key, ""); err != nil {
+			t.Fatal(err)
+		}
+
+		bp = getBasePath(ctx, o, root, key)
+
+		expected = expected[2:]
+		expected[0] = "rw"
+
+		test.expUID = 0
+		test.expGID = 0
+
+		checkMountOpts()
 	}
+}
 
-	if sn, ok := o.(*snapshotter); !ok || !sn.remapIDs {
-		t.Skip("overlayfs doesn't support idmapped mounts")
-	}
+func testOverlayRemappedActive(t *testing.T, newSnapshotter testsuite.SnapshotterFunc) {
+	for _, test := range []struct {
+		idMap        userns.IDMap
+		snapOpt      func(idMap userns.IDMap) snapshots.Opt
+		expUID       uint32
+		expGID       uint32
+		expUIDMntOpt string
+		expGIDMntOpt string
+	}{
+		{
+			idMap: userns.IDMap{
+				UidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      666,
+						Size:        65536,
+					},
+				},
+				GidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      666,
+						Size:        65536,
+					},
+				},
+			},
+			snapOpt: func(idMap userns.IDMap) snapshots.Opt {
+				return containerd.WithRemapperLabels(
+					idMap.UidMap[0].ContainerID, idMap.UidMap[0].HostID,
+					idMap.GidMap[0].ContainerID, idMap.GidMap[0].HostID,
+					idMap.UidMap[0].Size,
+				)
+			},
+			expUID:       666,
+			expGID:       666,
+			expUIDMntOpt: "uidmap=0:666:65536",
+			expGIDMntOpt: "gidmap=0:666:65536",
+		},
+		{
+			idMap: userns.IDMap{
+				UidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      666,
+						Size:        1000,
+					},
+					{
+						ContainerID: 1000,
+						HostID:      6666,
+						Size:        64536,
+					},
+				},
+				GidMap: []specs.LinuxIDMapping{
+					{
+						ContainerID: 0,
+						HostID:      888,
+						Size:        1000,
+					},
+					{
+						ContainerID: 1000,
+						HostID:      8888,
+						Size:        64536,
+					},
+				},
+			},
+			snapOpt: func(idMap userns.IDMap) snapshots.Opt {
+				return containerd.WithUserNSRemapperLabels(idMap.UidMap, idMap.GidMap)
+			},
+			expUID:       666,
+			expGID:       888,
+			expUIDMntOpt: "uidmap=0:666:1000,1000:6666:64536",
+			expGIDMntOpt: "gidmap=0:888:1000,1000:8888:64536",
+		},
+	} {
+		var (
+			opts   []snapshots.Opt
+			mounts []mount.Mount
+		)
 
-	hostID := uint32(666)
-	contID := uint32(0)
-	length := uint32(65536)
+		ctx := context.TODO()
+		root := t.TempDir()
+		o, _, err := newSnapshotter(ctx, root)
+		if err != nil {
+			t.Fatal(err)
+		}
 
-	uidMap := specs.LinuxIDMapping{
-		ContainerID: contID,
-		HostID:      hostID,
-		Size:        length,
-	}
-	gidMap := specs.LinuxIDMapping{
-		ContainerID: contID,
-		HostID:      hostID,
-		Size:        length,
-	}
-	opts = append(opts, containerd.WithRemapperLabels(
-		uidMap.ContainerID, uidMap.HostID,
-		gidMap.ContainerID, gidMap.HostID,
-		length),
-	)
+		if sn, ok := o.(*snapshotter); !ok || !sn.remapIDs {
+			t.Skip("overlayfs doesn't support idmapped mounts")
+		}
 
-	key := "/tmp/test"
-	if mounts, err = o.Prepare(ctx, key, "", opts...); err != nil {
-		t.Fatal(err)
-	}
+		opts = append(opts, test.snapOpt(test.idMap))
 
-	bp := getBasePath(ctx, o, root, key)
-	expected := []string{
-		fmt.Sprintf("uidmap=%d:%d:%d", uidMap.ContainerID, uidMap.HostID, uidMap.Size),
-		fmt.Sprintf("gidmap=%d:%d:%d", gidMap.ContainerID, gidMap.HostID, gidMap.Size),
-		"rw",
-		"rbind",
-	}
+		key := "/tmp/test"
+		if _, err = o.Prepare(ctx, key, "", opts...); err != nil {
+			t.Fatal(err)
+		}
+		if err = o.Commit(ctx, "base", key, opts...); err != nil {
+			t.Fatal(err)
+		}
+		if mounts, err = o.Prepare(ctx, key, "base", opts...); err != nil {
+			t.Fatal(err)
+		}
 
-	checkMountOpts := func() {
 		if len(mounts) != 1 {
 			t.Errorf("should only have 1 mount but received %d", len(mounts))
 		}
 
-		if len(mounts[0].Options) != len(expected) {
-			t.Errorf("expected %d options, but received %d", len(expected), len(mounts[0].Options))
+		bp := getBasePath(ctx, o, root, key)
+		expected := []string{
+			test.expUIDMntOpt, test.expGIDMntOpt,
+			fmt.Sprintf("workdir=%s", filepath.Join(bp, "work")),
+			fmt.Sprintf("upperdir=%s", filepath.Join(bp, "fs")),
+			fmt.Sprintf("lowerdir=%s", getParents(ctx, o, root, key)[0]),
 		}
 
 		m := mounts[0]
 		for i, v := range expected {
 			if m.Options[i] != v {
-				t.Errorf("mount option %q is not valid, expected %q", m.Options[i], v)
+				t.Errorf("mount option %q is invalid, expected %q", m.Options[i], v)
 			}
 		}
 
@@ -268,118 +540,12 @@ func testOverlayRemappedBind(t *testing.T, newSnapshotter testsuite.SnapshotterF
 		if err != nil {
 			t.Errorf("failed to stat %s", filepath.Join(bp, "fs"))
 		}
-
 		if stat, ok := st.Sys().(*syscall.Stat_t); !ok {
 			t.Errorf("incompatible types after stat call: *syscall.Stat_t expected")
-		} else if stat.Uid != uidMap.HostID || stat.Gid != gidMap.HostID {
-			t.Errorf("bad mapping: expected {uid: %d, gid: %d}; real {uid: %d, gid: %d}", uidMap.HostID, gidMap.HostID, int(stat.Uid), int(stat.Gid))
+		} else if stat.Uid != test.expUID || stat.Gid != test.expGID {
+			t.Errorf("bad mapping: expected {uid: %d, gid: %d}; received {uid: %d, gid: %d}", test.expUID, test.expGID, int(stat.Uid), int(stat.Gid))
 		}
 	}
-	checkMountOpts()
-
-	expected[2] = "ro"
-	if err = o.Commit(ctx, "base", key, opts...); err != nil {
-		t.Fatal(err)
-	}
-	if mounts, err = o.View(ctx, key, "base", opts...); err != nil {
-		t.Fatal(err)
-	}
-	bp = getBasePath(ctx, o, root, key)
-	checkMountOpts()
-
-	key = "/tmp/test1"
-	if mounts, err = o.Prepare(ctx, key, ""); err != nil {
-		t.Fatal(err)
-	}
-
-	bp = getBasePath(ctx, o, root, key)
-
-	expected = expected[2:]
-	expected[0] = "rw"
-
-	uidMap.HostID = 0
-	gidMap.HostID = 0
-
-	checkMountOpts()
-}
-
-func testOverlayRemappedActive(t *testing.T, newSnapshotter testsuite.SnapshotterFunc) {
-	var (
-		opts   []snapshots.Opt
-		mounts []mount.Mount
-	)
-
-	ctx := context.TODO()
-	root := t.TempDir()
-	o, _, err := newSnapshotter(ctx, root)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	if sn, ok := o.(*snapshotter); !ok || !sn.remapIDs {
-		t.Skip("overlayfs doesn't support idmapped mounts")
-	}
-
-	hostID := uint32(666)
-	contID := uint32(0)
-	length := uint32(65536)
-
-	uidMap := specs.LinuxIDMapping{
-		ContainerID: contID,
-		HostID:      hostID,
-		Size:        length,
-	}
-	gidMap := specs.LinuxIDMapping{
-		ContainerID: contID,
-		HostID:      hostID,
-		Size:        length,
-	}
-	opts = append(opts, containerd.WithRemapperLabels(
-		uidMap.ContainerID, uidMap.HostID,
-		gidMap.ContainerID, gidMap.HostID,
-		length),
-	)
-
-	key := "/tmp/test"
-	if _, err = o.Prepare(ctx, key, "", opts...); err != nil {
-		t.Fatal(err)
-	}
-	if err = o.Commit(ctx, "base", key, opts...); err != nil {
-		t.Fatal(err)
-	}
-	if mounts, err = o.Prepare(ctx, key, "base", opts...); err != nil {
-		t.Fatal(err)
-	}
-
-	if len(mounts) != 1 {
-		t.Errorf("should only have 1 mount but received %d", len(mounts))
-	}
-
-	bp := getBasePath(ctx, o, root, key)
-	expected := []string{
-		fmt.Sprintf("uidmap=%d:%d:%d", uidMap.ContainerID, uidMap.HostID, uidMap.Size),
-		fmt.Sprintf("gidmap=%d:%d:%d", gidMap.ContainerID, gidMap.HostID, gidMap.Size),
-		fmt.Sprintf("workdir=%s", filepath.Join(bp, "work")),
-		fmt.Sprintf("upperdir=%s", filepath.Join(bp, "fs")),
-		fmt.Sprintf("lowerdir=%s", getParents(ctx, o, root, key)[0]),
-	}
-
-	m := mounts[0]
-	for i, v := range expected {
-		if m.Options[i] != v {
-			t.Errorf("mount option %q is invalid, expected %q", m.Options[i], v)
-		}
-	}
-
-	st, err := os.Stat(filepath.Join(bp, "fs"))
-	if err != nil {
-		t.Errorf("failed to stat %s", filepath.Join(bp, "fs"))
-	}
-	if stat, ok := st.Sys().(*syscall.Stat_t); !ok {
-		t.Errorf("incompatible types after stat call: *syscall.Stat_t expected")
-	} else if stat.Uid != uidMap.HostID || stat.Gid != gidMap.HostID {
-		t.Errorf("bad mapping: expected {uid: %d, gid: %d}; received {uid: %d, gid: %d}", uidMap.HostID, gidMap.HostID, int(stat.Uid), int(stat.Gid))
-	}
 }
 
 func testOverlayRemappedInvalidMapping(t *testing.T, newSnapshotter testsuite.SnapshotterFunc) {
@@ -414,6 +580,24 @@ func testOverlayRemappedInvalidMapping(t *testing.T, newSnapshotter testsuite.Sn
 				snapshots.LabelSnapshotGIDMapping: "-666:-666:-666",
 			}),
 		},
+		"WithLabels: negative UID in multiple mappings must fail": {
+			snapshots.WithLabels(map[string]string{
+				snapshots.LabelSnapshotUIDMapping: "1:1:1,-1:-1:-2",
+				snapshots.LabelSnapshotGIDMapping: "0:0:66666",
+			}),
+		},
+		"WithLabels: negative GID in multiple mappings must fail": {
+			snapshots.WithLabels(map[string]string{
+				snapshots.LabelSnapshotUIDMapping: "0:0:66666",
+				snapshots.LabelSnapshotGIDMapping: "-1:-1:-2,6:6:6",
+			}),
+		},
+		"WithLabels: negative GID/UID in multiple mappings must fail": {
+			snapshots.WithLabels(map[string]string{
+				snapshots.LabelSnapshotUIDMapping: "-666:-666:-666,1:1:1",
+				snapshots.LabelSnapshotGIDMapping: "-666:-666:-666,2:2:2",
+			}),
+		},
 		"WithRemapperLabels: container ID (GID/UID) other than 0 must fail": {
 			containerd.WithRemapperLabels(666, 666, 666, 666, 666),
 		},
@@ -423,6 +607,24 @@ func testOverlayRemappedInvalidMapping(t *testing.T, newSnapshotter testsuite.Sn
 		"WithRemapperLabels: container ID (GID) other than 0 must fail": {
 			containerd.WithRemapperLabels(0, 0, 666, 0, 4294967295),
 		},
+		"WithUserNSRemapperLabels: container ID (GID/UID) other than 0 must fail": {
+			containerd.WithUserNSRemapperLabels(
+				[]specs.LinuxIDMapping{{ContainerID: 666, HostID: 666, Size: 666}},
+				[]specs.LinuxIDMapping{{ContainerID: 666, HostID: 666, Size: 666}},
+			),
+		},
+		"WithUserNSRemapperLabels: container ID (UID) other than 0 must fail": {
+			containerd.WithUserNSRemapperLabels(
+				[]specs.LinuxIDMapping{{ContainerID: 666, HostID: 0, Size: 65536}},
+				[]specs.LinuxIDMapping{{ContainerID: 0, HostID: 0, Size: 65536}},
+			),
+		},
+		"WithUserNSRemapperLabels: container ID (GID) other than 0 must fail": {
+			containerd.WithUserNSRemapperLabels(
+				[]specs.LinuxIDMapping{{ContainerID: 0, HostID: 0, Size: 4294967295}},
+				[]specs.LinuxIDMapping{{ContainerID: 666, HostID: 0, Size: 4294967295}},
+			),
+		},
 	} {
 		t.Log(desc)
 		if _, err = o.Prepare(ctx, key, "", opts...); err == nil {
